from typing import List, Tuple

from blspy import PopSchemeMPL, G1Element, G2Element
from eth2deposit.utils.ssz import (
    compute_deposit_domain,
    DepositMessage,
    compute_signing_root,
)
from web3 import Web3

from lido.contract import LidoContract
from lido.methods.typing import OperatorKey
from lido.network.type import WITHDRAWAL_CREDENTIALS, GENESIS_FORK_VERSION


def find_keys_duplicates(
    keys: List[OperatorKey],
) -> List[Tuple[OperatorKey, OperatorKey]]:
    """
    Find all duplicates in list of keys

    @param keys: List of keys
    @return: Returns list of pair keys (key_1, key_2) that are duplicating
    """
    keys_dict = {}
    duplicates = []

    for key in keys:
        if key["key"] not in keys_dict:
            keys_dict[key["key"]] = key
        else:
            duplicates.append((key, keys_dict[key["key"]]))

    return duplicates


def _get_withdrawal_credentials(chain_id: int):
    """
    @param chain_id: Network chain id
    @return: Possible withdrawal credentials
    """
    return [bytes.fromhex(cred[2:]) for cred in WITHDRAWAL_CREDENTIALS[chain_id]]


def validate_keys(w3: Web3, keys: List[OperatorKey]) -> List[OperatorKey]:
    """
    @param w3: Web3
    @param keys: List of keys to validate
    @return: List of keys that are invalid
    """
    actual_credentials = LidoContract.getWithdrawalCredentials(w3)[""]
    withdrawal_credentials = [
        *_get_withdrawal_credentials(w3.eth.chain_id),
        actual_credentials,
    ]

    deposit_domain = compute_deposit_domain(GENESIS_FORK_VERSION[w3.eth.chain_id])

    invalid_keys = []

    for key in keys:
        is_valid = validate_key(key, withdrawal_credentials, deposit_domain)
        if not is_valid:
            invalid_keys.append(key)

    return invalid_keys


def validate_key(
    key: OperatorKey, withdrawal_credentials: List[bytes], deposit_domain: bytes
) -> bool:
    """
    @param key: Key to check.
    @param withdrawal_credentials: List of possible creds.
    @param deposit_domain: Magic bytes.
    @return: Bool - Valid or Invalid this key
    """
    g1_pub_key = G1Element.from_bytes(key["key"])
    g2_signature = G2Element.from_bytes(key["depositSignature"])

    for wc in withdrawal_credentials:
        ETH32 = 32 * 10 ** 9
        deposit_message = DepositMessage(
            pubkey=key["key"],
            withdrawal_credentials=wc,
            amount=ETH32,
        )

        message = compute_signing_root(deposit_message, deposit_domain)
        is_valid = PopSchemeMPL.verify(g1_pub_key, message, g2_signature)

        if is_valid:
            return is_valid

    return False



from concurrent.futures import ThreadPoolExecutor

from multicall import Call, Multicall as DefaultMulticall

from lido.multicall.multicall_address import MULTICALL_ADDRESSES


class Multicall(DefaultMulticall):
    """
    Upgraded version of Multicall from https://github.com/banteg/multicall.py

    Improves:
        - Added MAX_CALLS_PER_MULTICALL param to avoid huge and slow batches in Multicall
        - results from multicall is not a dict, but a list now. We are making a lot of requests to one contract's method,
        so we don't wanna loose data.
    """

    MAX_CALLS_PER_MULTICALL = 275
    MAX_WORKERS = 6

    def __call__(self):
        calls_list = [
            self.calls[i : i + self.MAX_CALLS_PER_MULTICALL]
            for i in range(0, len(self.calls), self.MAX_CALLS_PER_MULTICALL)
        ]

        with ThreadPoolExecutor(max_workers=self.MAX_WORKERS) as executor:
            thread_results = executor.map(self.execute, calls_list)

        result = []

        for thread_result in thread_results:
            result.extend(thread_result)

        return result

    def execute(self, calls):
        aggregate = Call(
            MULTICALL_ADDRESSES[self.w3.eth.chainId],
            "aggregate((address,bytes)[])(uint256,bytes[])",
            returns=None,
            _w3=self.w3,
            block_id=self.block_id,
        )

        args = [[[call.target, call.data] for call in calls]]
        block, outputs = aggregate(args)

        results = []
        for call, output in zip(self.calls, outputs):
            results.append(call.decode_output(output))

        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command


# Package meta-data.
NAME = "lido-sdk"
DESCRIPTION = (
    "This library consolidates various functions to efficiently load network data for Lido,"
    " validate node operator keys and find key duplicates."
)
URL = "https://github.com/lidofinance/lido-validator-python"
EMAIL = "info@lido.fi"
AUTHOR = "Lido"
REQUIRES_PYTHON = ">=3.7,<4"
VERSION = "0.2.2"

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class UploadCommand(Command):
    """
    Support setup.py upload.
        Note: To use the 'upload' functionality of this file, you must:
        $ pip install twine
    """

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    # If your package is a single module, use this instead of "packages":
    # py_modules=["mypackage"],
    # entry_points={
    #     "console_scripts": ["mycli=mymodule:cli"],
    # },
    install_requires=[
        'multicall==0.1.2',
        'web3==5.23.0',
        'blspy==1.0.5',
        'eth2deposit==1.2.0',
        'ssz==0.2.4',
    ],
    tests_require=['pytest==6.2.4'],
    include_package_data=True,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
    dependency_links=[
        'http://github.com/ethereum/eth2.0-deposit-cli/tarball/v1.2.0#egg=eth2_deposit-1.2.0',
    ],
)
